/*
 * lirc_client.c - Client library for communicating with the LIRC
 * client daemon
 * Copyright (C) 2003  Fredrik Tolf (fredrik@dolda2000.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>
#include <errno.h>
#include <regex.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <sys/uio.h>
#include <sys/wait.h>

#include "lirc_client.h"

#define EXEC_REGEX "cd:[[:blank:]]*(([^#]|\\\\#)+)(#|$)"
#define DEFAULT_CD "/usr/local/bin/lirccdd:/usr/bin/lirccd"
#define DEBUG

static int beverbose = 0;
static int sockfd = -1;
static char *inbuf = NULL;
static int bufsize = 0, datalen = 0;
static int warning = 1;

static char *quoteword(char *word)
{
    int i;
    char *buf;
    int bufsize, len;
    
    if((buf = malloc(bufsize = ((len = (strlen(word) + 1)) + 16))) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "liblirc_client: quoteword: out of memory\n");
	errno = ENOMEM;
	return(NULL);
    }
    strcpy(buf, word);
    for(i = 0; buf[i]; i++)
    {
	if((buf[i] == ' ') || (buf[i] == '\\'))
	{
	    if(++len >= bufsize)
	    {
		if((buf = realloc(buf, bufsize += 16)) == NULL)
		{
		    if(beverbose)
			fprintf(stderr, "liblirc_client: quoteword: out of memory\n");
		    free(buf);
		    errno = ENOMEM;
		    return(NULL);
		}
	    }
	    memmove(buf + i + 1, buf + i, len - i);
	    buf[i++] = '\\';
	}
    }
    return(buf);
}

static char *wordcmp(char **words)
{
    char *buf, *buf2;
    int len;
    
    if((buf = malloc(1)) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "liblirc_client: wordcmp: out of memory\n");
	errno = ENOMEM;
	return(NULL);
    }
    *buf = 0;
    len = 1;
    for(; *words; words++)
    {
	buf2 = quoteword(*words);
	if((buf = realloc(buf, len += (strlen(buf2) + 1))) == NULL)
	{
	    if(beverbose)
		fprintf(stderr, "liblirc_client: wordcmp: out of memory\n");
	    free(buf);
	    free(buf2);
	    errno = ENOMEM;
	    return(NULL);
	}
	if(*buf)
	    strcat(buf, " ");
	strcat(buf, buf2);
	free(buf2);
    }
    return(buf);
}

static char **wordexp(char *line)
{
    char **words;
    char *buf, *p;
    int bufsize, len, numwords;
    
    if((words = malloc(sizeof(char *))) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
	errno = ENOMEM;
	return(NULL);
    }
    *words = NULL;
    numwords = 0;
    if((buf = malloc(bufsize = 16)) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
	free(words);
	errno = ENOMEM;
	return(NULL);
    }
    *buf = 0;
    len = 0;
    for(p = line; 1; p++)
    {
	if(!*p || (*p == ' '))
	{
	    buf[len++] = 0;
	    buf = realloc(buf, len);
	    words[numwords++] = buf;
	    if((words = realloc(words, sizeof(char *) * (numwords + 1))) == NULL)
	    {
		if(beverbose)
		    fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
		free(words);
		free(buf);
		errno = ENOMEM;
		return(NULL);
	    }
	    words[numwords] = NULL;
	    if((buf = malloc(bufsize = 16)) == NULL)
	    {
		if(beverbose)
		    fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
		free(words);
		errno = ENOMEM;
		return(NULL);
	    }
	    *buf = 0;
	    len = 0;
	    if(!*p)
		break;
	} else if(*p == '\\') {
	    if(*(p + 1))
	    {
		buf[len++] = *(++p);
		if(len == bufsize)
		{
		    if((buf = realloc(buf, bufsize += 16)) == NULL)
		    {
			if(beverbose)
			    fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
			free(words);
			free(buf);
			errno = ENOMEM;
			return(NULL);
		    }
		}
	    }
	} else {
	    buf[len++] = *p;
	    if(len == bufsize)
	    {
		if((buf = realloc(buf, bufsize += 16)) == NULL)
		{
		    if(beverbose)
			fprintf(stderr, "liblirc_client: wordexp: out of memory\n");
		    free(words);
		    free(buf);
		    errno = ENOMEM;
		    return(NULL);
		}
	    }
	}
    }
    return(words);
}

void lirc_freewords(char **words)
{
    char **p;
    
    for(p = words; *p; p++)
	free(*p);
    free(words);
}

static int opensock(char *path)
{
    struct sockaddr_un name;
    int fd;
    
    name.sun_family = AF_UNIX;
    strcpy(name.sun_path, path);
    if((fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
	return(-1);
    if(connect(fd, (struct sockaddr *)&name, sizeof(name)) < 0)
    {
	close(fd);
	return(-1);
    }
    return(fd);
}

static int getinput(int block)
{
    struct pollfd pfd;
    int ret;
    
    if(sockfd < 0)
    {
	errno = EBADF;
	return(-1);
    }
    pfd.fd = sockfd;
    pfd.events = POLLIN;
    poll(&pfd, 1, block?-1:0);
    if(pfd.revents)
    {
	if(bufsize - datalen < 64)
	{
	    if((inbuf = realloc(inbuf, bufsize += 128)) == NULL)
	    {
		if(beverbose)
		    fprintf(stderr, "liblirc_client: getinput: out of memory\n");
		errno = ENOMEM;
		return(-1);
	    }
	}
	ret = read(sockfd, inbuf + datalen, bufsize - datalen);
	if(ret < 0)
	{
	    if(beverbose)
		fprintf(stderr, "liblirc_client: getinput: %s\n", strerror(errno));
	    return(-1);
	}
	if(ret == 0)
	{
	    if(beverbose)
		fprintf(stderr, "liblirc_client: getinput: client daemon closed connection\n");
	    errno = EPIPE;
	    return(-1);
	}
	datalen += ret;
	return(ret);
    } else {
	return(0);
    }
}

static char *getline(char *type, int block)
{
    int i, o, t;
    char *n, *buf;
    
    t = 0;
    i = 0;
    while(1)
    {
	while(1)
	{
	    if((n = (char *)memchr(inbuf + i, '\n', datalen - i)) == NULL)
		break;
	    o = (n - inbuf) + 1;
	    if(!memcmp(type, inbuf + i, strlen(type)))
	    {
		buf = malloc(o - i);
		memcpy(buf, inbuf + i, (o - i) - 1);
		buf[(o - i) - 1] = 0;
		memmove(inbuf + o, inbuf + i, (datalen -= (o - i)) - i);
		return(buf);
	    }
	    i = o;
	}
	if(block)
	{
	    getinput(1);
	} else {
	    if(!t)
	    {
		getinput(0);
	    } else {
		errno = EAGAIN;
		return(NULL);
	    }
	    t = 1;
	}
    }
}

static int sendcmd(char **words)
{
    int ret;
    struct iovec v[2];
    unsigned char *buf;
    
    buf = wordcmp(words);
    v[0].iov_base = buf;
    v[0].iov_len = strlen(buf);
    v[1].iov_base = "\n";
    v[1].iov_len = 1;
    /*
     * Use writev in order to avoid having to realloc buf just to
     * strcat a simple \n.
     */
    ret = writev(sockfd, v, 2);
    free(buf);
    return(ret);
}

int lirc_sendcmd(char *cmd, ...)
{
    int i;
    char **words;
    int bufsize, ret;
    va_list args;
    
    if(sockfd < 0)
    {
	errno = EBADF;
	return(-1);
    }
    if((words = malloc(sizeof(char *) * (bufsize = 8))) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "lirc_sendcmd: out of memory\n");
	errno = ENOMEM;
	return(-1);
    }
    words[0] = cmd;
    va_start(args, cmd);
    while(1)
    {
	if((words[i++] = va_arg(args, char *)) == NULL)
	    break;
	if((i > bufsize) && ((words = realloc(words, sizeof(char *) * (bufsize += 8))) == NULL))
	{
	    free(words);
	    if(beverbose)
		fprintf(stderr, "lirc_sendcmd: out of memory\n");
	    errno = ENOMEM;
	    return(-1);
	}
    }
    va_end(args);
    ret = sendcmd(words);
    free(words);
    return(ret);
}

int lirc_getreply(char ***words, int block)
{
    char **w, *l;
    int i, ret;
    
    if((l = getline("repl", block)) == NULL)
	return(-1);
    w = wordexp(l);
    free(l);
    if(w == NULL)
	return(-1);
    ret = atoi(w[1]);
    if(words == NULL)
    {
	lirc_freewords(w);
    } else {
	free(w[0]);
	free(w[1]);
	i = 2;
	do
	{
	    w[i - 2] = w[i];
	    i++;
	} while(w[i] != NULL);
	*words = w;
    }
    return(ret);
}

int lirc_getmesg(char **mesg, int block)
{
    char **w, **wp, *l;
    
    if((l = getline("mesg", block)) == NULL)
    {
	if(mesg != NULL)
	    *mesg = NULL;
	return(-1);
    }
    w = wordexp(l);
    free(l);
    if(w == NULL)
    {
	if(mesg != NULL)
	    *mesg = NULL;
	return(-1);
    }
    if(mesg != NULL)
	*mesg = w[1];
    for(wp = w; *wp; wp++)
    {
	if(wp != w + 1)
	    free(*wp);
    }
    free(w);
    return(0);
}

int lirc_getbutton(char **code, int *repeat, char **button, char **remote, int block)
{
    char **w, **wp, *l;
    
    if((l = getline("bttn", block)) == NULL)
    {
	errno = EAGAIN;
	return(-1);
    }
    w = wordexp(l);
    free(l);
    if(w == NULL)
	return(-1);
    free(w[0]);
    if(code != NULL)
	*code = w[1];
    else
	free(w[1]);
    if(repeat != NULL)
	*repeat = atoi(w[2]);
    free(w[2]);
    if(button != NULL)
	*button = w[3];
    else
	free(w[3]);
    if(remote != NULL)
	*remote = w[4];
    else
	free(w[4]);
    /* Not that there _should_ be more words, but what do you know? */
    for(wp = w + 5; *wp; wp++)
	free(*wp);
    free(w);
    return(0);
}

int lirc_runcmd(char ***w, char *cmd, ...)
{
    int i;
    char **words;
    int bufsize, ret;
    va_list args;
    
    if(sockfd < 0)
    {
	errno = EBADF;
	return(-1);
    }
    if((words = malloc(sizeof(char *) * (bufsize = 8))) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "lirc_sendcmd: out of memory\n");
	errno = ENOMEM;
	return(-1);
    }
    words[0] = cmd;
    va_start(args, cmd);
    i = 1;
    while(1)
    {
	if((words[i++] = va_arg(args, char *)) == NULL)
	    break;
	if((i > bufsize) && ((words = realloc(words, sizeof(char *) * (bufsize += 8))) == NULL))
	{
	    free(words);
	    if(beverbose)
		fprintf(stderr, "lirc_sendcmd: out of memory\n");
	    errno = ENOMEM;
	    return(-1);
	}
    }
    va_end(args);
    ret = sendcmd(words);
    free(words);
    if(ret < 0)
	return(ret);
    return(lirc_getreply(w, 1));
}

static int startcd(void)
{
    int i, fd;
    char *homedir, *buf, *p, *p2;
    FILE *rcfile;
    regex_t preg;
    regmatch_t matches[16];
    int ret;
    pid_t pid;
    
    if((homedir = getenv("HOME")) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "startcd: $HOME not set\n");
	errno = EINVAL;
	return(-1);
    }
    if((buf = (char *)malloc(strlen(homedir) + 9)) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "startcd: no memory\n");
	return(-1);
    }
    strcpy(buf, homedir);
    strcat(buf, "/.lircrc");
    if(((rcfile = fopen(buf, "r")) == NULL) && beverbose)
	perror(buf);
    free(buf);
    if(rcfile == NULL)
	return(-1);
    if((buf = malloc(128)) == NULL)
    {
	if(beverbose)
	    fprintf(stderr, "startcd: no memory\n");
	return(-1);
    }
    fgets(buf, 128, rcfile);
    if((p = strchr(buf, '\n')) != NULL)
	*p = 0;
    if(ret = regcomp(&preg, EXEC_REGEX, REG_EXTENDED))
    {
	if(beverbose)
	    fprintf(stderr, "startcd: internal error, invalid regex\n");
#ifdef DEBUG
	regerror(ret, &preg, buf, 128);
	fprintf(stderr, "EXEC_REGEX (" EXEC_REGEX ") error: %s\n", buf);
#endif
	free(buf);
	fclose(rcfile);
	return(-1);
    }
    if(!regexec(&preg, buf, 16, matches, 0))
    {
	memmove(buf, buf + matches[1].rm_so, matches[1].rm_eo - matches[1].rm_so);
	buf[matches[1].rm_eo - matches[1].rm_so] = 0;
    } else {
	strcpy(buf, DEFAULT_CD);
	p = strtok_r(buf, ":", &p2);
	do
	{
	    if(!access(p, X_OK))
		break;
	} while((p = strtok_r(NULL, ":", &p2)) != NULL);
	if(p == NULL)
	{
	    if(beverbose)
		fprintf(stderr, "startcd: could not find a client daemon\n");
	    regfree(&preg);
	    free(buf);
	    fclose(rcfile);
	    return(-1);
	}
	memmove(buf, p, strlen(p) + 1);
    }
    regfree(&preg);
    fclose(rcfile);
    if(!(pid = fork()))
    {
	if(beverbose)
	{
	    for(i = 3; i < 1024; i++)
		close(i);
	} else {
	    for(i = 0; i < 1024; i++)
		close(i);
	    fd = open("/dev/null", O_RDWR);
	    dup2(fd, 0);
	    dup2(fd, 1);
	    dup2(fd, 2);
	    if(fd > 2)
		close(fd);
	}
	execl(buf, buf, "-c", NULL);
	if(beverbose)
	    fprintf(stderr, "startcd: could not exec %s, %s\n", buf, strerror(errno));
	exit(127);
    }
    if(pid < 0)
    {
	if(beverbose)
	    fprintf(stderr, "startcd: could not fork: %s\n", strerror(errno));
	free(buf);
	return(-1);
    }
    waitpid(pid, &ret, 0);
    free(buf);
    return(ret);
}

int lirc_init2(char *prog, int verbose, char *sock, ...)
{
    char *homedir, *buf;
    int fd;
    int tries;
    
    beverbose = verbose;
    if((inbuf == NULL) && ((inbuf = malloc(bufsize = 128)) == NULL))
    {
	if(beverbose)
	    fprintf(stderr, "lirc_init2: out of memory\n");
	errno = ENOMEM;
	return(-1);
    }
    tries = 0;
    while(1)
    {
	if(sock == NULL)
	{
	    if((homedir = getenv("HOME")) == NULL)
	    {
		if(beverbose)
		    fprintf(stderr, "lirc_init: $HOME not set\n");
		errno = EINVAL;
		return(-1);
	    }
	    if((buf = (char *)malloc(strlen(homedir) + 8)) == NULL)
	    {
		if(beverbose)
		    fprintf(stderr, "lirc_init: no memory\n");
		return(-1);
	    }
	    strcpy(buf, homedir);
	    strcat(buf, "/.lircc");
	    fd = opensock(buf);
	    free(buf);
	} else {
	    fd = opensock(sock);
	}
	if(fd < 0)
	{
	    if((errno == ECONNREFUSED) || (errno == ENOENT))
	    {
		if(tries++ < 3)
		{
		    if(!startcd())
			continue;
		    else
			return(-1);
		} else {
		    if(beverbose)
			fprintf(stderr, "lirc_init: "
				"cannot get client daemon running, "
				"giving up!\n");
		    return(-1);
		}
	    }
	    if(beverbose)
	    {
		fprintf(stderr, "lirc_init: could not open socket to client daemon: %s\n", strerror(errno));
		return(-1);
	    }
	}
	break;
    }
    sockfd = fd;
    if((prog != NULL) && LIRCC_FAILED(lirc_runcmd(NULL, "ident", prog, NULL)))
    {
	close(fd);
	sockfd = -1;
	errno = EBADMSG; /* ? */
	return(-1);
    }
    return(fd);
}

int lirc_deinit(void)
{
    if(inbuf != NULL)
    {
	free(inbuf);
	bufsize = 0;
    }
    if(sockfd >= 0)
	return(close(sockfd));
    return(0);
}

/* 
 * The following are compatibility calls.
 *
 * Their original semantics and the implementation in this library are
 * documented for each call, so that any misunderstandings can be
 * spotted and corrected.
 */

#if 1
#define COMP do { if(beverbose && warning) \
    { \
	fprintf(stderr, \
		"This program is using the obsolete lirc calls.\n" \
		"Please contact the author and tell him/her to use" \
		"the newer calls instead\n"); \
        warning = 0; \
    } } while(0)
#else
#define COMP
#endif

/*
 * lirc_init:
 *  Call semantics:
 *   Called with program id and verbosity mode.
 *   Returns
 *    -1 Failure (errno set)
 *    0  Success
 *  Implementation:
 *   lirc_init2 behaves the same with regard to program id, verbosity
 *   mode and return value, so the sockname and extra parameters are
 *   passed as NULL, which should give original behavior.
 */

int lirc_init(char *prog, int verbose)
{
    COMP;
    return(lirc_init2(prog, verbose, NULL, NULL));
}

/*
 * lirc_nextcode:
 *  Call semantics:
 *   Called with a pointer to a buffer which shall contain the
 *   received string. Blocking mode is set using the O_NONBLOCK file
 *   descriptor flag.
 *   Returns:
 *    -1 Failure (errno set)
 *    0  Success
 *    0  O_NONBLOCK was set and no code was available (errno = EAGAIN)
 *  Implementation:
 *   O_NONBLOCK is checked with fcntl and passed to lirc_getmesg for
 *   obtaining the right block behavior. Also, if lirc_getmesg returns
 *   -1, errno is checked for EAGAIN, since lirc_getmesg considers that
 *   en error (for consistency with other syscalls). Otherwise,
 *   lirc_getmesg already behaves the same.
 *   Note: lirc_nextcode is supposed to return the "physical" remote
 *   control code as given by lircd. However, lirccd doesn't allow
 *   retrieving raw codes except when explicitly asked for. Since
 *   there is no way of predicting what the program wants, the final
 *   message is always returned and lirc_code2char is implemented as a
 *   noop, since this is probably the behavior that most applications
 *   want. Programs that want to parse raw codes will simply have to
 *   be rewritten.
 */

int lirc_nextcode(char **code)
{
    int ret;
    
    COMP;
    ret = lirc_getmesg(code, !(fcntl(sockfd, F_GETFL) & (O_NONBLOCK | O_ASYNC)));
    if((ret < 0) && (errno == EAGAIN))
	ret = 0;
    return(ret);
}

/*
 * lirc_code2char:
 *  Call semantics:
 *   Called with a pointer to the config data, a buffer containing the
 *   raw code and a pointer to a buffer for holding the translated
 *   string. This is supposed to be a static buffer (actually,
 *   contained in the config data). From what I can see, the client is
 *   supposed to loop this call until the returned string is NULL. I'm
 *   guessing it is in the case that there are multiple matching
 *   configuration entries.
 *   Returns:
 *    -1                     Failure
 *    0 and *string != NULL  Success
 *    0 and *string == NULL  Syntax error in the raw code, or end of
 *                           possible translations.
 *  Implementation:
 *   Since code will already be a translated string, it is simply
 *   copied to a static buffer (so that the existing memory management
 *   logic will continue to work). To accomplish the looping
 *   functionality, I have implemented a static variable (to avoid
 *   having to actually allocate a config buffer in lirc_readconfig).
 */

int lirc_code2char(struct lirc_config *config, char *code, char **string)
{
    static char compbuf[1024];
    static int loop = 1;
    
    COMP;
    if(string == NULL)
	return(0);
    if(!loop)
    {
	*string = NULL;
	loop = 1;
	return(0);
    }
    if(strlen(code) < 1024)
    {
	strcpy(compbuf, code);
    } else {
	memcpy(compbuf, code, 1023);
	compbuf[1023] = 0;
    }
    *string = compbuf;
    loop = 0;
    return(0);
}

/*
 * lirc_readconfig:
 *  Call semantics:
 *   Called with a buffer containing the config file path, or NULL for
 *   defaulting to $HOME/.lircrc, a pointer to a buffer to hold the
 *   parsed config data, and a function which I'm not sure what it's
 *   supposed to do. I guess programs that depend on it won't work as
 *   expected.
 *   Returns:
 *    -1 Failure (errno set)
 *    0  Success
 *  Implementation:
 *   Since there is no concept like this with lirccd, this is a
 *   noop. It sets the config data pointer to NULL and returns
 *   success. Programs that require the exact behavior offered by
 *   this, such as returning a correct config data pointer, will have
 *   to be rewritten.
 */

int lirc_readconfig(char *file, struct lirc_config **config, int (check)(char *s))
{
    COMP;
    /* Should there be a mandatory warning if the program passes a
       non-NULL "check"? */
    if(config != NULL)
	*config = NULL;
    return(0);
}

/*
 * lirc_freeconfig:
 *  Call semantics:
 *   Called with a config data pointer previously obtained with
 *   lirc_readconfig.
 *   Does not return a value.
 *  Implementation:
 *   Noop
 */

void lirc_freeconfig(struct lirc_config *config)
{
    COMP;
}

/*
 * lirc_nextir:
 * This call was already obsoleted before lirccd.
 *  Call semantics:
 *   No arguments.
 *   Returns:
 *    NULL           Failure (errno set)
 *    Valid pointer  Pointer to the read code is returned; ie. success
 *  Implementation:
 *   Precisely like before. This function is copied almost verbatim
 *   from the old lirc_client.c, except the reference to lirc_prog is
 *   removed, and the global warning flag is used instead.
 */

char *lirc_nextir(void)
{
	char *code;
	int ret;
	
	if(warning)
	{
	    fprintf(stderr, "warning: lirc_nextir() is _really_ obsolete\n");
	    COMP;
	}
	ret = lirc_nextcode(&code);
	if(ret == -1)
	    return(NULL);
	return(code);
}

/*
 * lirc_ir2char:
 * This call was already obsoleted before lirccd.
 *  Call semantics:
 *   Called with a pointer to the parsed config data and a buffer
 *   holding the raw code.
 *   Returns:
 *    NULL           Failure (errno set)
 *    Valid pointer  Pointer to the translated string is returned;
 *                   ie. success
 *  Implementation:
 *   Copied from the old lirc_client.c with the same modifications as
 *   lirc_nextir.
 */

char *lirc_ir2char(struct lirc_config *config, char *code)
{
	char *string;
	
	if(warning)
	{
	    fprintf(stderr, "warning: lirc_ir2char() is _really_ obsolete\n");
	    COMP;
	}
	if(lirc_code2char(config, code, &string) == -1)
	    return(NULL);
	return(string);
}
